use crate::*;
use serde::{Deserialize, Serialize};

const APP_NAME: &str = "gcli";

/// defines structure of config file
#[derive(Serialize, Deserialize, Debug)]
pub struct Config {
	// duniter endpoint
	pub duniter_endpoint: String,
	// indexer endpoint
	pub indexer_endpoint: String,
	// user address
	pub address: Option<AccountId>,
	// user secret (substrate format)
	pub secret: Option<String>,
}

impl std::default::Default for Config {
	fn default() -> Self {
		Self {
			duniter_endpoint: String::from(data::LOCAL_DUNITER_ENDPOINT),
			indexer_endpoint: String::from(data::LOCAL_INDEXER_ENDPOINT),
			address: None,
			secret: None,
		}
	}
}

/// load config file and manage error if could not
pub fn load_conf() -> Config {
	match confy::load(APP_NAME, None) {
		Ok(cfg) => cfg,
		Err(e) => {
			log::warn!(
				"met error while loading config file {}",
				confy::get_configuration_file_path(APP_NAME, None)
					.unwrap()
					.display()
			);
			log::error!("{:?}", e);
			log::info!("using default config instead");
			log::info!("call `config save` to overwrite");
			Config::default()
		}
	}
}

#[derive(Clone, Default, Debug, clap::Parser)]
pub enum Subcommand {
	#[default]
	/// Show config path
	Where,
	/// Show config
	Show,
	/// Save config as modified by command line arguments
	Save,
}

/// handle conf command
pub fn handle_command(data: Data, command: Subcommand) -> anyhow::Result<()> {
	// match subcommand
	match command {
		Subcommand::Where => {
			println!(
				"{}",
				confy::get_configuration_file_path(APP_NAME, None)?.display()
			);
		}
		Subcommand::Show => {
			println!("{:?}", data.cfg);
		}
		Subcommand::Save => {
			confy::store(APP_NAME, None, &data.cfg).expect("unable to write config");
		}
	};

	Ok(())
}
