use crate::*;

use clap::builder::OsStr;
use std::str::FromStr;

// #[derive(Clone, Copy, Debug, Eq, PartialEq)]
// pub enum NeededKeys {
// 	None,
// 	Public,
// 	Secret,
// }

#[derive(Clone, Copy, Debug, Eq, PartialEq, Default)]
pub enum SecretFormat {
	/// Raw 32B seed
	Seed,
	/// Substrate secret key or BIP39 mnemonic (optionally followed by derivation path)
	#[default]
	Substrate,
	/// Predefined (Alice, Bob, ...)
	Predefined,
}

impl FromStr for SecretFormat {
	type Err = std::io::Error;

	fn from_str(s: &str) -> std::io::Result<Self> {
		match s {
			"seed" => Ok(SecretFormat::Seed),
			"substrate" => Ok(SecretFormat::Substrate),
			"predefined" => Ok(SecretFormat::Predefined),
			_ => Err(std::io::Error::from(std::io::ErrorKind::InvalidInput)),
		}
	}
}

impl From<SecretFormat> for &'static str {
	fn from(val: SecretFormat) -> &'static str {
		match val {
			SecretFormat::Seed => "seed",
			SecretFormat::Substrate => "substrate",
			SecretFormat::Predefined => "predefined",
		}
	}
}

impl From<SecretFormat> for OsStr {
	fn from(val: SecretFormat) -> OsStr {
		OsStr::from(Into::<&str>::into(val))
	}
}

/// get keypair from given string secret
pub fn pair_from_str(secret_format: SecretFormat, secret: &str) -> anyhow::Result<Pair> {
	match secret_format {
		SecretFormat::Seed => {
			let mut seed = [0; 32];
			hex::decode_to_slice(secret, &mut seed).map_err(|_| anyhow!("Invalid secret"))?;
			let pair = Pair::from_seed(&seed);
			Ok(pair)
		}
		// "predefined" replaces secret before
		SecretFormat::Substrate | SecretFormat::Predefined => {
			Pair::from_string(secret, None).map_err(|_| anyhow!("Invalid secret"))
		}
	}
}

/// get keypair and address from given secret string
pub fn addr_and_pair_from_secret(
	secret_format: SecretFormat,
	secret: &str,
) -> anyhow::Result<(AccountId, Pair)> {
	let pair = pair_from_str(secret_format, secret)?;
	let address = pair.public().into();
	Ok((address, pair))
}

/// ask user to input a secret
pub fn prompt_secret(secret_format: SecretFormat) -> Pair {
	loop {
		match pair_from_str(
			secret_format,
			&rpassword::prompt_password(format!("Secret key ({secret_format:?}): ")).unwrap(),
		) {
			Ok(pair) => return pair,
			Err(_) => println!("Invalid secret"),
		}
	}
}

// /// get keys with interactive prompt if needed
// pub fn get_keys(
// 	secret_format: SecretFormat,
// 	accout_id: Option<AccountId>,
// 	secret: &Option<String>,
// 	needed_keys: NeededKeys,
// ) -> anyhow::Result<(Option<AccountId>, Option<Pair>)> {
// 	// Get from args
// 	let mut account_id = match (accout_id, secret) {
// 		// both are defined, check that they match.
// 		// if they do not match, use secret
// 		(Some(accout_id), Some(secret)) => {
// 			let pair = pair_from_str(secret_format, secret)?;
// 			let secret_address = pair.public().into();
// 			if accout_id != secret_address {
// 				println!("Secret ({secret_address}) and address ({accout_id}) do not match, using secret");
// 			}
// 			return Ok((Some(secret_address), Some(pair)));
// 		}
// 		// only secret, build both
// 		(None, Some(secret)) => {
// 			let pair = pair_from_str(secret_format, secret)?;
// 			return Ok((Some(pair.public().into()), Some(pair)));
// 		}
// 		// only address
// 		(Some(accout_id), None) => Some(accout_id),
// 		// none of them
// 		(None, None) => None,
// 	};

// 	// Prompt
// 	if needed_keys == NeededKeys::Secret
// 		|| (account_id.is_none() && needed_keys == NeededKeys::Public)
// 	{
// 		loop {
// 			let pair = prompt_secret(secret_format);

// 			if let Some(account_id) = &account_id {
// 				if account_id != &pair.public().into() {
// 					println!("Secret and address do not match.");
// 				}
// 			} else {
// 				account_id = Some(pair.public().into());
// 				return Ok((account_id, Some(pair)));
// 			}
// 		}
// 	}

// 	Ok((account_id, None))
// }
